/*
  t2moon

  Copyright 2010 Jere Leppnen <jere.leppanen@gmail.com>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.  
*/

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <libgen.h>

#define LOG(level, args...) do { if(par.verbose>level) { printf("%s(%d): ", __FUNCTION__, __LINE__); printf(args); printf("\n"); } } while(0)
#define ERROR(args...) do { printf("%s(%d): ERROR: ", __FUNCTION__, __LINE__); printf(args); printf("\n"); exit(1); } while(0)

#define PROGRAM "t2moon"
#define NEEDLE "CELOBJVAR"

struct
{
  int verbose;
  char *file;
} par={0, NULL};

int fileexists(const char *name)
{
  LOG(2, " ");
  FILE *file=fopen(name, "rb");
  if(file) fclose(file);
  LOG(1, "File %s %s.", name, file ? "exists" : "doesn't exist");
  return file!=NULL;
}

void searchdir()
{
  LOG(2, " ");

  int c;
  char *cdir="", *ldir=NULL;
  for(c=0; c<128; c++)
    {
      free(ldir);
      ldir=cdir;
      if(!(cdir=getcwd(NULL, 0))) ERROR("getcwd()");
      LOG(3, "%s", cdir);
      if(!strcmp(ldir, cdir)) break;
      if(fileexists("thief2.exe"))
	{
	  printf("Thief2 directory is %s.\n", cdir);
	  return;
	}
      chdir("..");
    }
  ERROR("Cannot find Thief2 installation directory.");
}

int confirm(const char *msg)
{
  LOG(2, " ");

  printf(msg);
  fflush(stdout);
  char tmp[128];
  memset(tmp, 0, 128);
  scanf("%s", tmp);
  if(*tmp=='y' || *tmp=='Y') return 1;
  return 0;
}

void cmd_help()
{
  LOG(2, " ");
  printf("\nUsage: %s [command] [options]\n"
	 "\nCommands:\n\n"
	 "    misinfo   Show the current values.\n"
         "              Nothing to do with misinformation.\n\n"
	 "    mispatch  Patch the values in *.mis file(s).\n\n"
	 "\nOptions:\n"
	 "    -f <file>   operate on specified file\n"
	 "    -h          help\n"
	 "    -v <level>  verbosity level\n",
	 PROGRAM);
  printf("\n\
This program replaces the stock moon in Thief 2: The Metal Age with a\n\
more realistic one.  This is a texture replacement, but also the\n\
color, alpha, rotation and phase are modified.\n\
\n\
Requires ddfix (http://timeslip.chorrol.com/ddfix.html) with\n\
Thief2Extensions on.\n\
\n\
The easiest way to use this program is to place it anywhere under your\n\
Thief2 installation directory and then just double-click on\n\
t2moon.exe. You'll be prompted on each file that is patched.\n\
\n\
Only a single command may be specified. If no command is given, the\n\
default is to execute mispatch. Options may appear in any order, the\n\
last of conflicting options takes effect.\n\
\n\
Thief2 installation directory is searched by starting at the current\n\
working directory and going up the directory tree level by level until\n\
thief2.exe is found. Any relative paths used with the -f option are\n\
relative to this directory.\n\
\n");
  exit(1);
}

struct
{
  char name[9];
  char num;
  char unused[14];
  int enabled, fog, alphatex;
  char texture[256];
  float alpha, offset, size, latitude, longitude, rotation;
  float r, g, b;
} celobj;

struct
{
  char *texture;
  float size, rotation;
  float r, g, b;
} celobj_miss[]={
  // 1 Running Interference
  // CELOBJVAR1: tex=moon07a a=1 siz=10 lat=65 lon=110 rot=120 col=(0.8 0.59 0.13)
#define AVG ((.8+.59+.13)/3.0)
#define W1 3.0
#define TW 4.0
  { "pmoon08.pcx", 8, 220, (.8*W1+AVG)/TW, (.59*W1+AVG)/TW, (.13*W1+AVG)/TW },
  // 2 Shipping and Receiving
  // CELOBJVAR3: tex=moon08a.tga a=1 siz=10 lat=30 lon=45 rot=120 col=(1 1 1)
  { "pmoon09.pcx", 6, 240, 1, 1, 1 },
  // skip
  { NULL },
  // 3 Framed
  // CELOBJVAR1: tex=moon10a.tga a=0.7 siz=10 lat=55 lon=65 rot=120 col=(0.9 0.9 1)
  //{ "pmoon05.pcx", 7, 230, .9, .9, 1 },
  { "pmoonfr.pcx", 7, 230, .9, .9, 1 },
  // 4 Ambush!
  // CELOBJVAR1: tex=moon13a.tga a=0.7 siz=10 lat=45 lon=30 rot=120 col=(0.8 0.8 0.5)
  { "pmoon06.pcx", 7, 250, .8*.7, .8*.7, .5*.7 },
  // 5 Eavesdropping
  // CELOBJVAR1: tex=moon13a.tga a=0.7 siz=10 lat=30 lon=90 rot=120 col=(0.9 0.9 0.7)
  { "pmoon06.pcx", 6, 270, .9*.7, .9*.7, .7*.7 },
  // 6 First City Bank and Trust
  // CELOBJVAR1: tex=moon13a.tga a=1 siz=12 lat=30 lon=30 rot=120 col=(1 1 1)
  { "pmoon07.pcx", 6, 250, 1, 1, 1 },
  // 8 Trace the Courier
  // CELOBJVAR1: tex=moon16a.tga a=1 siz=10 lat=45 lon=90 rot=120 col=(1 1 1)
  { "pmoon09.pcx", 7, 210, 1, 1, 1 },
  // 7 Blackmail
  // CELOBJVAR1: tex=moon14a.tga a=0.8 siz=7 lat=50 lon=50 rot=120 col=(0.9 0.9 1)
  { "pmoon08.pcx", 7, 200, .9*.8, .9*.8, .8 },
  // 9 Trail of Blood
  // CELOBJVAR1: tex=moon16a.tga a=1 siz=10 lat=45 lon=90 rot=120 col=(1 1 1)
  { "pmoon09.pcx", 7, 210, 1, 1, 1 },
  // 10 Life of the Party
  // CELOBJVAR1: tex=moon03a.tga a=1 siz=10 lat=30 lon=270 rot=120 col=(1 1 1)
  { "pmoon03.pcx", 6, 340, 1, 1, 1 },
  // 13 Casing The Joint
  // CELOBJVAR1: tex=moon06a.tga a=1 siz=10 lat=45 lon=180 rot=120 col=(0.65 0.65 0.65)
  { "pmoon05.pcx", 7, 330, .65, .65, .65 },
  // 14 Masks
  // CELOBJVAR1: tex=moon10a.tga a=0.25 siz=10 lat=45 lon=180 rot=120 col=(1 0 0)
  { "pmoonec.pcx", 7, 330, .3, .3, .3 },
  // 11 Precious Cargo
  { NULL },
  // 12 Kidnap
  { NULL },
  // 15 Sabotage at Soulforge
  // CELOBJVAR1: tex=moon11a.tga a=0.7 siz=10 lat=45 lon=90 rot=120 col=(0.9 0.9 1)
  { "pmoon08.pcx", 7, 220, .9*.7, .9*.7, .7 }
};

void print_celobj()
{
  printf("%s%d: ena=%d fog=%d iat=%d tex=%s a=%g off=%g siz=%g lat=%g lon=%g rot=%g col=(%g %g %g)\n", NEEDLE, (int)celobj.num-(int)'0', celobj.enabled, celobj.fog, celobj.alphatex, celobj.texture, celobj.alpha, celobj.offset, celobj.size, celobj.latitude, celobj.longitude, celobj.rotation, celobj.r, celobj.g, celobj.b);
}

#define BUFSIZE 4096

int misinfo(const char *name)
{
  printf("\nmisfile=%s\n\n", name);

  size_t offset=0;
  
  FILE *file=fopen(name, "rb");
  if(!file) ERROR("Cannot open file %s.", name);
  
  char *buf=malloc(BUFSIZE*2);
  memset(buf, 0, BUFSIZE*2);

  const char *needle=NEEDLE;

  int block=-1;
  while(!feof(file))
    {
      memcpy(buf, buf+BUFSIZE, BUFSIZE);
      int i=fread(buf+BUFSIZE, 1, BUFSIZE, file);
      memset(buf+BUFSIZE+i, 0, BUFSIZE-i);
      char *p=memchr(buf, *needle, BUFSIZE);
      while(p)
	{
	  if(!memcmp(needle, p, strlen(needle)))
	    {
	      offset=block*BUFSIZE+(p-buf);
	      LOG(1, "Found %s at 0x%x", needle, offset);
	      memcpy(&celobj, p, sizeof(celobj));
	      if(celobj.enabled==1) print_celobj();
	    }
	  p=memchr(p+1, *needle, BUFSIZE-(p-buf)-1);
	}
      block++;
    }
  fclose(file);
  return 0;
}

void cmd_misinfo()
{
  LOG(2, " ");

  if(par.file)
    {
      if(misinfo(par.file)) ERROR("%s doesn't look like a mis file.", par.file);
      return;
    }

  DIR *dir=opendir(".");
  if(!dir) ERROR("Cannot open current directory.");
  struct dirent *ent;
  while((ent=readdir(dir)))
    {
      LOG(3, "dirent=%s", ent->d_name);
      if(!strcasecmp(ent->d_name+strlen(ent->d_name)-4, ".mis"))
	misinfo(ent->d_name);
    }
}

void mispatch(const char *name)
{
  LOG(2, "name=%s", name);

  size_t offset=0;
  
  FILE *file=fopen(name, "rb");
  if(!file) 
    {      
      LOG(0, "Cannot open file %s.", name);
      return;
    }
  
  char *p=(char *)name+strlen(name)-1;
  while(p>=name && *p!='\\' && *p!='/') p--;
  int num=0;
  if(sscanf(p+1, "miss%d", &num)<1)
    {
      LOG(0, "Cannot get mission number for file %s.", name);
      return;
    }
  LOG(0, "Mission number %d. sizeof=%d", num, sizeof(celobj_miss)/sizeof(celobj_miss[0]));
  if(num<1 || num>sizeof(celobj_miss)/sizeof(celobj_miss[0]) || !celobj_miss[num-1].texture)
    {
      LOG(0, "Don't know how to patch mission number %d.", num);
      return;
    }
  num--;
  
  char *buf=malloc(BUFSIZE*2);
  memset(buf, 0, BUFSIZE*2);

  const char *needle=NEEDLE;

  int block=-1;
  while(!feof(file))
    {
      memcpy(buf, buf+BUFSIZE, BUFSIZE);
      int i=fread(buf+BUFSIZE, 1, BUFSIZE, file);
      memset(buf+BUFSIZE+i, 0, BUFSIZE-i);
      p=memchr(buf, *needle, BUFSIZE);
      while(p)
	{
	  if(!memcmp(needle, p, strlen(needle)))
	    {
	      offset=block*BUFSIZE+(p-buf);
	      LOG(1, "Found %s at 0x%x", needle, offset);
	      memcpy(&celobj, p, sizeof(celobj));
	      if(celobj.enabled==1) 
		{
		  fclose(file);
		  printf("\nPatching file %s:\n", name);
		  printf("Old values:\n");
		  print_celobj();
		  printf("New values:\n");
		  memset(celobj.texture, 0, 256);
		  strcpy(celobj.texture, celobj_miss[num].texture);
		  celobj.size=celobj_miss[num].size;
		  celobj.rotation=celobj_miss[num].rotation;
		  celobj.r=celobj_miss[num].r;
		  celobj.g=celobj_miss[num].g;
		  celobj.b=celobj_miss[num].b;
		  celobj.alpha=1;
		  celobj.fog=1;
		  print_celobj();
		  if(confirm("\nPatching is not undoable. Information will be lost.\n"
			     "Are you sure you've made backups and want to continue (y/n)? "))
		    {
		      FILE *file=fopen(name, "r+b");
		      if(!file) ERROR("Cannot open %s for writing.", name);
		      if(fseek(file, offset, SEEK_SET)) ERROR("Cannot seek to 0x%x in %s.", offset, name);
		      if(fwrite(&celobj, sizeof(celobj), 1, file)!=1) ERROR("Cannot write to %s.", name);
		      fclose(file);
		      printf("%s patched.\n", name);
		      return;
		    }
		  printf("No changes made to %s.\n", name);
		  return;		  
		}
	    }
	  p=memchr(p+1, *needle, BUFSIZE-(p-buf)-1);
	}
      block++;
    }
  fclose(file);
  LOG(0, "I don't know what to do with file %s.", name);
  return;
}

void cmd_mispatch()
{
  LOG(2, " ");

  if(par.file)
    {
      mispatch(par.file);
      return;
    }

  mispatch("miss1.mis");
  mispatch("miss2.mis");
  mispatch("miss4.mis");
  mispatch("miss5.mis");
  mispatch("miss6.mis");
  mispatch("miss7.mis");
  mispatch("miss8.mis");
  mispatch("miss9.mis");
  mispatch("miss10.mis");
  mispatch("miss11.mis");
  mispatch("miss12.mis");
  mispatch("miss13.mis");
  mispatch("miss16.mis");
}

void chkw()
{
  LOG(2, " ");

  if(sizeof(float)!=4)
    ERROR("sizeof(float)=%d != 4", sizeof(float));
}

int main(int argc, char *argv[])
{
  int c;
  for(c=1; c<argc; c++)
    {
      if(!strncasecmp(argv[c], "-h", 2)) cmd_help();
    }

  opterr=0;
  while(1)
    {
      int o=getopt(argc, argv, "f:v:");
      if(o<0) break;
      switch(o)
	{
	case 'f':
	  par.file=strdup(optarg);
	  break;
	case 'v':
	  par.verbose=atoi(optarg);
	  break;
	case ':': case '?': default:
	  ERROR("Invalid option \'%c\' or missing value. (Try %s -h)", optopt, PROGRAM);
	}
    }

  LOG(1, "%s %s", __FILE__, __DATE__);

  chkw();

  searchdir();

  if(optind<argc)
    {
      if(!strcasecmp(argv[optind], "misinfo")) cmd_misinfo();
      else if(!strcasecmp(argv[optind], "mispatch")) cmd_mispatch();
      else ERROR("Invalid command %s. (Try %s -h)", argv[optind], PROGRAM);
      return 0;
    }

  cmd_mispatch();

  return 0;
}
